/* pitwrite: write binary pit from pitdump output */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <getopt.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include "partition.h"

#define RESERVE_INIT_SIZE 16
#define MAX_TOKEN 12
#define MAX_PART 256

struct token_map {
     unsigned int offset;
     char *name;
};

#define PI_TOKEN(name) { offsetof(partition_info, name), #name }


void die(int use_perror, char *msg, ...) __attribute__((noreturn));
void die(int use_perror, char *msg, ...)
{
     va_list ap;
     static char buf[512];

     va_start(ap, msg);
     vsnprintf(buf, sizeof(buf), msg, ap);
     va_end(ap);

     if (use_perror)
	  perror(buf);
     else
	  fprintf(stderr, "%s\n", buf);

     exit(1);
}

/* make sure there is at least elem_size*count bytes available */
void *reserve(void *ptr, size_t elem_size, unsigned int count, unsigned int *size)
{
     if (!ptr) {
	  *size = RESERVE_INIT_SIZE;
	  ptr = malloc(elem_size * (*size));
     } else if (count >= *size) {
	  int sz = *size * 2;
	  *size *= sz < count ? count : sz;
	  ptr = realloc(ptr, elem_size * (*size));
     }

     if (!ptr)
	  die(0, "Unable to allocate memory");

     return ptr;
}

void get_token(char *buf, unsigned int token, partition_info *pi)
{
     char *fmt;

     static struct token_map tok[MAX_TOKEN] = {
	  PI_TOKEN(szName),
	  PI_TOKEN(szFileName),
	  PI_TOKEN(szDeltaName),
	  PI_TOKEN(nBinType),
	  PI_TOKEN(nDevType),
	  PI_TOKEN(nID),
	  PI_TOKEN(nAttribute),
	  PI_TOKEN(nUpdateAttr),
	  PI_TOKEN(dwBlkSize),
	  PI_TOKEN(dwBlkLen),
	  PI_TOKEN(dwOffset),
	  PI_TOKEN(dwFileSize),
     };

     if (token >= 0 && token < 3)
	  fmt = "%31s";
     else if (token >= 3 && token < 8)
	  fmt = "%d";
     else if (token < 12)
	  fmt = "%u";
     else
	  die(0, "Internal error - trying to token out of allowed range");

     if (sscanf(buf, fmt, (void *)pi + tok[token].offset) != 1
	 && token != 2) /* szDelta name can be empty */
	  die (0, "Unable to parse token %d: %s - got %s",
	       token, tok[token].name, buf);
}

int main(int argc, char **argv)
{
     int token, loop, c;
     char *p = NULL;            /* buffer for tokens */
     unsigned int plen = 0;
     unsigned int psz = 0;
     partition_header ph;
     partition_info *pi = NULL;
     unsigned int off = 0;      /* offset */
     unsigned int sz = 0;       /* total size */
     FILE *fin = stdin;
     FILE *fout = stdout;

     while ((c = getopt(argc, argv, "i:o:")) != -1) {
	  switch (c) {
	  case 'i':
	       fin = fopen(optarg, "r");
	       if (!fin)
		    die(1, "Unable to open input file %s", optarg);
	       break;

	  case 'o':
	       fout = fopen(optarg, "w");
	       if (!fout)
		    die(1, "Unable to open output file %s", optarg);
	       break;

	  default:
	       die(0, "usage: %s [-i INPUT] [-o OUTPUT]]", argv[0]);
	  }
     }

     if (fout == stdout && isatty(1))
	  die(0, "I won't write binary data to terminal device");

     memset(&ph, 0, sizeof(ph));

     /* skip header */
     while(fgetc(fin) != '\n')
	  ;

     for (loop = 1 ; loop; )  {
	  pi = reserve(pi, sizeof(*pi), off, &sz);

	  memset(&pi[off], 0, sizeof(*pi));

	  for (plen = token = 0; token < MAX_TOKEN; ) {
	       int c = fgetc(fin);

	       p = reserve(p, 1, plen, &psz);
	       if (c == EOF) {
		    loop = 0;
		    break;
	       } else if (c == ',' || c == '\n') {
		    p[plen] = 0;
		    get_token(p, token, &pi[off]);
		    plen = 0;
		    p[plen] = 0;
		    token++;
	       } else {
		    p[plen] = c;
		    plen++;
	       }
	  }
	  off++;
     }

     off -= 1;
     if (off == 0 || off > MAX_PART)
	  die(0, "Invalid number of partitions");

     ph.dwMagic = PARTITION_MAGIC;
     ph.nCount = off;

     fwrite(&ph, sizeof(ph), 1, fout);
     fwrite(pi, sizeof(partition_info), off, fout);
     if (fflush(fout))
	  die(1, "Error while writing output file");

     return 0;
}
